//   Copyright (c) Gerard Lanois, 2001.
//
//   This program is freely distributable without licensing fees and is
//   provided without guarantee or warrantee expressed or implied. This
//   program is -not- in the public domain. 

#include "ExtrudedString.h"
#include "FTGeometryGlyph.h"

#include <string.h>   // strlen()
#include <stdlib.h>   // exit()
#include <iostream.h> // cerr


ExtrudedString::ExtrudedString(
   const char* font_name, 
   const char* text)
{
   _font_name = 0;
   _text = 0;
   _text_length = 0;

   if (!font_name || !strlen(font_name) || !text || !strlen(text)) {
      cerr << "ExtrudedString::ExtrudedString() ERROR - ";
      cerr << "bad argument(s).\n";
      exit(1);
   }
   else {

      _font_name = new char[strlen(font_name)+1];
      if (!_font_name) {
         cerr << "ExtrudedString::ExtrudedString() ERROR - ";
         cerr << "new char[] failed.\n";
         exit(1);
      }
      else {
         strcpy(_font_name, font_name);

         _text_length = strlen(text);
         _text = new char[_text_length+1];
         if (!_text) {
            cerr << "ExtrudedString::ExtrudedString() ERROR - ";
            cerr << "new char[] failed.\n";
            exit(1);
         }
         else {
            strcpy(_text, text);

            if (!_font.Open(_font_name)) {
                cerr << "ExtrudedString::ExtrudedString() ERROR - ";
                cerr << "unable to open font" << _font_name << "\n";
            }
            else {
                int point_size = 24;
                if (!_font.FaceSize(point_size)) {
                    cerr << "ERROR: Unable to set font face size " << point_size << "\n";
                }
            }
         }
      }
   }
}


ExtrudedString::~ExtrudedString()
{
    if (_font_name) {
        delete[] _font_name;
    }
    if (_text) {
        delete[] _text;
    }
}


void
ExtrudedString::draw(GLdouble extrude_length)
{
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix(); {

        // Draw extruded string.
        FTFace* face = _font.GetFace();
        vector<FTGeometryGlyph*>* glyphs = _font.GetGlyphList();
        double base_x = 0.0;
        for (int i=0; i<_text_length; ++i) {
            unsigned int left = face->CharIndex( _text[i]);
            FTGeometryGlyph* glyph = (*glyphs)[left];

            // Do all the triangles (front and back).
            double* triangles = glyph->triangles;
            int nTriangles = glyph->numTriangles;
            glBegin(GL_TRIANGLES); {
                
                for (int j=0; j<nTriangles/3; ++j) {

                    double* p1= triangles + j*9;
                    double* p2= triangles + j*9 + 3;
                    double* p3= triangles + j*9 + 6;

                    // Do the front facing triangle.
                    glNormal3d(0.0, 0.0, 1.0);
                    glVertex3d(p3[0]+base_x, p3[1], 0.0);
                    glVertex3d(p2[0]+base_x, p2[1], 0.0);
                    glVertex3d(p1[0]+base_x, p1[1], 0.0);
                    
                    // Do the back facing triangle.
                    glNormal3d(0.0, 0.0, -1.0);
                    glVertex3d(p1[0]+base_x, p1[1], -extrude_length);
                    glVertex3d(p2[0]+base_x, p2[1], -extrude_length);
                    glVertex3d(p3[0]+base_x, p3[1], -extrude_length);
                } // for
            } glEnd();

            // Extrude each contour to make the sides.
            double* contour = glyph->contours;
            for (int c=0; c<glyph->numContours; ++c) {
            
                // Make a quad strip using each successive
                // pair of points in this contour.
                int numPoints = glyph->contourLength[c];
                glBegin(GL_QUAD_STRIP); {
                    for (int j= 0; j <= numPoints; ++j) {
                        int j1 = (j < numPoints) ? j : 0;
                        int j0 = (j1 == 0) ? (numPoints-1) : (j1-1);

                        double* p0 = contour + j0*3;
                        double* p1 = contour + j1*3;

                        // Compute normal for this quad.
                        double vx = p1[0] - p0[0];
                        double vy = p1[1] - p0[1];
                        glNormal3f(-vy, vx, 0.0);
                        
                        // Add vertices to the quad strip.
                        glVertex3f(p0[0]+base_x, p0[1], -extrude_length);
                        glVertex3f(p0[0]+base_x, p0[1], 0.0);
                    } // for
                } glEnd();
                contour += numPoints*3;
            } // for 

            // Do kerning.
            FT_Vector kernAdvance;	
            kernAdvance.x = 0; 
            kernAdvance.y = 0;
            if (i < _text_length-1) {
                unsigned int right = face->CharIndex( _text[i+1]);
                kernAdvance = face->KernAdvance( left, right);
            }
            float advance = 0.0;
            if( !face->Error()) {
                advance = glyph->Advance();
            }
            kernAdvance.x += (FT_Pos)advance;
            base_x += kernAdvance.x;
        }

    } glPopMatrix();
}


void
ExtrudedString::getBBox( 
   int& llx, 
   int& lly, 
   int& urx, 
   int& ury )
{
    llx = 0;
    lly = (int)_font.Descender();
    urx = (int)_font.Advance(_text);
    ury = (int)_font.Ascender();
}


